@@ -92,7 +92,7 @@ class Cache {
 	void swap_index(int i, int j)
 	{
 		if(i==j) return;
-		
+
 		if(head[i].len > 0) lru_delete(head[i]);
 		if(head[j].len > 0) lru_delete(head[j]);
 		do {float[] tmp=head[i].data; head[i].data=head[j].data; head[j].data=tmp;} while(false);
@@ -179,7 +179,7 @@ abstract class Kernel extends QMatrix {
 			case svm_parameter.PRECOMPUTED:
 				return x[i][(int)(x[j][0].value)].value;
 			default:
-				return 0;	// java
+				return 0;	// Unreachable
 		}
 	}
 
@@ -198,7 +198,8 @@ abstract class Kernel extends QMatrix {
 			for(int i=0;i<l;i++)
 				x_square[i] = dot(x[i],x[i]);
 		}
-		else x_square = null;
+		else
+			x_square = null;
 	}
 
 	static double dot(svm_node[] x, svm_node[] y)
@@ -274,10 +275,10 @@ abstract class Kernel extends QMatrix {
 			}
 			case svm_parameter.SIGMOID:
 				return Math.tanh(param.gamma*dot(x,y)+param.coef0);
-			case svm_parameter.PRECOMPUTED:
+			case svm_parameter.PRECOMPUTED:  //x: test (validation), y: SV
 				return	x[(int)(y[0].value)].value;
 			default:
-				return 0;	// java
+				return 0;	// Unreachable
 		}
 	}
 }
@@ -318,7 +319,7 @@ class Solver {
 	double[] G_bar;		// gradient, if we treat free variables as 0
 	int l;
 	boolean unshrink;	// XXX
-	
+
 	static final double INF = java.lang.Double.POSITIVE_INFINITY;
 
 	double get_C(int i)
@@ -335,7 +336,7 @@ class Solver {
 	}
 	boolean is_upper_bound(int i) { return alpha_status[i] == UPPER_BOUND; }
 	boolean is_lower_bound(int i) { return alpha_status[i] == LOWER_BOUND; }
-	boolean is_free(int i) {  return alpha_status[i] == FREE; }
+	boolean is_free(int i) { return alpha_status[i] == FREE; }
 
 	// java: information about solution except alpha,
 	// because we cannot return multiple values otherwise...
@@ -386,7 +387,7 @@ class Solver {
 				for(j=0;j<active_size;j++)
 					if(is_free(j))
 						G[i] += alpha[j] * Q_i[j];
-			}	
+			}
 		}
 		else
 		{
@@ -484,7 +485,7 @@ class Solver {
 				else
 					counter = 1;	// do shrinking next iteration
 			}
-			
+
 			int i = working_set[0];
 			int j = working_set[1];
 
@@ -510,7 +511,7 @@ class Solver {
 				double diff = alpha[i] - alpha[j];
 				alpha[i] += delta;
 				alpha[j] += delta;
-			
+
 				if(diff > 0)
 				{
 					if(alpha[j] < 0)
@@ -630,7 +631,7 @@ class Solver {
 			}
 
 		}
-		
+
 		if(iter >= max_iter)
 		{
 			if(active_size < l)
@@ -674,18 +675,18 @@ class Solver {
 	{
 		// return i,j such that
 		// i: maximizes -y_i * grad(f)_i, i in I_up(\alpha)
-		// j: mimimizes the decrease of obj value
+		// j: minimizes the decrease of obj value
 		//    (if quadratic coefficeint <= 0, replace it with tau)
 		//    -y_j*grad(f)_j < -y_i*grad(f)_i, j in I_low(\alpha)
-		
+
 		double Gmax = -INF;
 		double Gmax2 = -INF;
 		int Gmax_idx = -1;
 		int Gmin_idx = -1;
 		double obj_diff_min = INF;
-	
+
 		for(int t=0;t<active_size;t++)
-			if(y[t]==+1)	
+			if(y[t]==+1)
 			{
 				if(!is_upper_bound(t))
 					if(-G[t] >= Gmax)
@@ -703,12 +704,12 @@ class Solver {
 						Gmax_idx = t;
 					}
 			}
-	
+
 		int i = Gmax_idx;
 		float[] Q_i = null;
 		if(i != -1) // null Q_i not accessed: Gmax=-INF if i=-1
 			Q_i = Q.get_Q(i,active_size);
-	
+
 		for(int j=0;j<active_size;j++)
 		{
 			if(y[j]==+1)
@@ -720,13 +721,13 @@ class Solver {
 						Gmax2 = G[j];
 					if (grad_diff > 0)
 					{
-						double obj_diff; 
+						double obj_diff;
 						double quad_coef = QD[i]+QD[j]-2.0*y[i]*Q_i[j];
 						if (quad_coef > 0)
 							obj_diff = -(grad_diff*grad_diff)/quad_coef;
 						else
 							obj_diff = -(grad_diff*grad_diff)/1e-12;
-	
+
 						if (obj_diff <= obj_diff_min)
 						{
 							Gmin_idx=j;
@@ -744,13 +745,13 @@ class Solver {
 						Gmax2 = -G[j];
 					if (grad_diff > 0)
 					{
-						double obj_diff; 
+						double obj_diff;
 						double quad_coef = QD[i]+QD[j]+2.0*y[i]*Q_i[j];
 						if (quad_coef > 0)
 							obj_diff = -(grad_diff*grad_diff)/quad_coef;
 						else
 							obj_diff = -(grad_diff*grad_diff)/1e-12;
-	
+
 						if (obj_diff <= obj_diff_min)
 						{
 							Gmin_idx=j;
@@ -770,7 +771,7 @@ class Solver {
 	}
 
 	private boolean be_shrunk(int i, double Gmax1, double Gmax2)
-	{	
+	{
 		if(is_upper_bound(i))
 		{
 			if(y[i]==+1)
@@ -782,7 +783,7 @@ class Solver {
 		{
 			if(y[i]==+1)
 				return(G[i] > Gmax2);
-			else	
+			else
 				return(G[i] > Gmax1);
 		}
 		else
@@ -800,7 +801,7 @@ class Solver {
 		{
 			if(y[i]==+1)
 			{
-				if(!is_upper_bound(i))	
+				if(!is_upper_bound(i))
 				{
 					if(-G[i] >= Gmax1)
 						Gmax1 = -G[i];
@@ -811,14 +812,14 @@ class Solver {
 						Gmax2 = G[i];
 				}
 			}
-			else		
+			else
 			{
-				if(!is_upper_bound(i))	
+				if(!is_upper_bound(i))
 				{
 					if(-G[i] >= Gmax2)
 						Gmax2 = -G[i];
 				}
-				if(!is_lower_bound(i))	
+				if(!is_lower_bound(i))
 				{
 					if(G[i] >= Gmax1)
 						Gmax1 = G[i];
@@ -826,11 +827,12 @@ class Solver {
 			}
 		}
 
-		if(unshrink == false && Gmax1 + Gmax2 <= eps*10) 
+		if(unshrink == false && Gmax1 + Gmax2 <= eps*10)
 		{
 			unshrink = true;
 			reconstruct_gradient();
 			active_size = l;
+			svm.info("*");
 		}
 
 		for(i=0;i<active_size;i++)
@@ -858,16 +860,16 @@ class Solver {
 		{
 			double yG = y[i]*G[i];
 
-			if(is_lower_bound(i))
+			if(is_upper_bound(i))
 			{
-				if(y[i] > 0)
+				if(y[i] < 0)
 					ub = Math.min(ub,yG);
 				else
 					lb = Math.max(lb,yG);
 			}
-			else if(is_upper_bound(i))
+			else if(is_lower_bound(i))
 			{
-				if(y[i] < 0)
+				if(y[i] > 0)
 					ub = Math.min(ub,yG);
 				else
 					lb = Math.max(lb,yG);
@@ -914,18 +916,18 @@ final class Solver_NU extends Solver
 		// j: minimizes the decrease of obj value
 		//    (if quadratic coefficeint <= 0, replace it with tau)
 		//    -y_j*grad(f)_j < -y_i*grad(f)_i, j in I_low(\alpha)
-	
+
 		double Gmaxp = -INF;
 		double Gmaxp2 = -INF;
 		int Gmaxp_idx = -1;
-	
+
 		double Gmaxn = -INF;
 		double Gmaxn2 = -INF;
 		int Gmaxn_idx = -1;
-	
+
 		int Gmin_idx = -1;
 		double obj_diff_min = INF;
-	
+
 		for(int t=0;t<active_size;t++)
 			if(y[t]==+1)
 			{
@@ -945,7 +947,7 @@ final class Solver_NU extends Solver
 						Gmaxn_idx = t;
 					}
 			}
-	
+
 		int ip = Gmaxp_idx;
 		int in = Gmaxn_idx;
 		float[] Q_ip = null;
@@ -954,25 +956,25 @@ final class Solver_NU extends Solver
 			Q_ip = Q.get_Q(ip,active_size);
 		if(in != -1)
 			Q_in = Q.get_Q(in,active_size);
-	
+
 		for(int j=0;j<active_size;j++)
 		{
 			if(y[j]==+1)
 			{
-				if (!is_lower_bound(j))	
+				if (!is_lower_bound(j))
 				{
 					double grad_diff=Gmaxp+G[j];
 					if (G[j] >= Gmaxp2)
 						Gmaxp2 = G[j];
 					if (grad_diff > 0)
 					{
-						double obj_diff; 
+						double obj_diff;
 						double quad_coef = QD[ip]+QD[j]-2*Q_ip[j];
 						if (quad_coef > 0)
 							obj_diff = -(grad_diff*grad_diff)/quad_coef;
 						else
 							obj_diff = -(grad_diff*grad_diff)/1e-12;
-	
+
 						if (obj_diff <= obj_diff_min)
 						{
 							Gmin_idx=j;
@@ -990,13 +992,13 @@ final class Solver_NU extends Solver
 						Gmaxn2 = -G[j];
 					if (grad_diff > 0)
 					{
-						double obj_diff; 
+						double obj_diff;
 						double quad_coef = QD[in]+QD[j]-2*Q_in[j];
 						if (quad_coef > 0)
 							obj_diff = -(grad_diff*grad_diff)/quad_coef;
 						else
 							obj_diff = -(grad_diff*grad_diff)/1e-12;
-	
+
 						if (obj_diff <= obj_diff_min)
 						{
 							Gmin_idx=j;
@@ -1009,13 +1011,13 @@ final class Solver_NU extends Solver
 
 		if(Math.max(Gmaxp+Gmaxp2,Gmaxn+Gmaxn2) < eps || Gmin_idx == -1)
 			return 1;
-	
+
 		if(y[Gmin_idx] == +1)
 			working_set[0] = Gmaxp_idx;
 		else
 			working_set[0] = Gmaxn_idx;
 		working_set[1] = Gmin_idx;
-	
+
 		return 0;
 	}
 
@@ -1025,14 +1027,14 @@ final class Solver_NU extends Solver
 		{
 			if(y[i]==+1)
 				return(-G[i] > Gmax1);
-			else	
+			else
 				return(-G[i] > Gmax4);
 		}
 		else if(is_lower_bound(i))
 		{
 			if(y[i]==+1)
 				return(G[i] > Gmax2);
-			else	
+			else
 				return(G[i] > Gmax3);
 		}
 		else
@@ -1045,7 +1047,7 @@ final class Solver_NU extends Solver
 		double Gmax2 = -INF;	// max { y_i * grad(f)_i | y_i = +1, i in I_low(\alpha) }
 		double Gmax3 = -INF;	// max { -y_i * grad(f)_i | y_i = -1, i in I_up(\alpha) }
 		double Gmax4 = -INF;	// max { y_i * grad(f)_i | y_i = -1, i in I_low(\alpha) }
- 
+
 		// find maximal violating pair first
 		int i;
 		for(i=0;i<active_size;i++)
@@ -1061,14 +1063,14 @@ final class Solver_NU extends Solver
 			if(!is_lower_bound(i))
 			{
 				if(y[i]==+1)
-				{	
+				{
 					if(G[i] > Gmax2) Gmax2 = G[i];
 				}
 				else	if(G[i] > Gmax3) Gmax3 = G[i];
 			}
 		}
 
-		if(unshrink == false && Math.max(Gmax1+Gmax2,Gmax3+Gmax4) <= eps*10) 
+		if(unshrink == false && Math.max(Gmax1+Gmax2,Gmax3+Gmax4) <= eps*10)
 		{
 			unshrink = true;
 			reconstruct_gradient();
@@ -1090,7 +1092,7 @@ final class Solver_NU extends Solver
 				}
 			}
 	}
-	
+
 	double calculate_rho()
 	{
 		int nr_free1 = 0,nr_free2 = 0;
@@ -1102,10 +1104,10 @@ final class Solver_NU extends Solver
 		{
 			if(y[i]==+1)
 			{
-				if(is_lower_bound(i))
-					ub1 = Math.min(ub1,G[i]);
-				else if(is_upper_bound(i))
+				if(is_upper_bound(i))
 					lb1 = Math.max(lb1,G[i]);
+				else if(is_lower_bound(i))
+					ub1 = Math.min(ub1,G[i]);
 				else
 				{
 					++nr_free1;
@@ -1114,10 +1116,10 @@ final class Solver_NU extends Solver
 			}
 			else
 			{
-				if(is_lower_bound(i))
-					ub2 = Math.min(ub2,G[i]);
-				else if(is_upper_bound(i))
+				if(is_upper_bound(i))
 					lb2 = Math.max(lb2,G[i]);
+				else if(is_lower_bound(i))
+					ub2 = Math.min(ub2,G[i]);
 				else
 				{
 					++nr_free2;
@@ -1293,7 +1295,7 @@ public class svm {
 	//
 	// construct and solve various formulations
 	//
-	public static final int LIBSVM_VERSION=322; 
+	public static final int LIBSVM_VERSION=332;
 	public static final Random rand = new Random();
 
 	private static svm_print_interface svm_print_stdout = new svm_print_interface()
@@ -1307,7 +1309,7 @@ public class svm {
 
 	private static svm_print_interface svm_print_string = svm_print_stdout;
 
-	static void info(String s) 
+	static void info(String s)
 	{
 		svm_print_string.print(s);
 	}
@@ -1471,7 +1473,7 @@ public class svm {
 		{
 			alpha2[i] = alpha2[i+l] = Math.min(sum,C);
 			sum -= alpha2[i];
-			
+
 			linear_term[i] = - prob.y[i];
 			y[i] = 1;
 
@@ -1484,7 +1486,7 @@ public class svm {
 			alpha2, C, C, param.eps, si, param.shrinking);
 
 		svm.info("epsilon = "+(-si.r)+"\n");
-		
+
 		for(i=0;i<l;i++)
 			alpha[i] = alpha2[i] - alpha2[i+l];
 	}
@@ -1495,7 +1497,7 @@ public class svm {
 	static class decision_function
 	{
 		double[] alpha;
-		double rho;	
+		double rho;
 	};
 
 	static decision_function svm_train_one(
@@ -1556,7 +1558,7 @@ public class svm {
 	}
 
 	// Platt's binary SVM Probablistic Output: an improvement from Lin et al.
-	private static void sigmoid_train(int l, double[] dec_values, double[] labels, 
+	private static void sigmoid_train(int l, double[] dec_values, double[] labels,
 				  double[] probAB)
 	{
 		double A, B;
@@ -1566,7 +1568,7 @@ public class svm {
 		for (i=0;i<l;i++)
 			if (labels[i] > 0) prior1+=1;
 			else prior0+=1;
-	
+
 		int max_iter=100;	// Maximal number of iterations
 		double min_step=1e-10;	// Minimal step taken in line search
 		double sigma=1e-12;	// For numerically strict PD of Hessian
@@ -1576,8 +1578,8 @@ public class svm {
 		double[] t= new double[l];
 		double fApB,p,q,h11,h22,h21,g1,g2,det,dA,dB,gd,stepsize;
 		double newA,newB,newf,d1,d2;
-		int iter; 
-	
+		int iter;
+
 		// Initial Point and Initial Fun Value
 		A=0.0; B=Math.log((prior0+1.0)/(prior1+1.0));
 		double fval = 0.0;
@@ -1623,7 +1625,7 @@ public class svm {
 			// Stopping Criteria
 			if (Math.abs(g1)<eps && Math.abs(g2)<eps)
 				break;
-			
+
 			// Finding Newton direction: -inv(H') * g
 			det=h11*h22-h21*h21;
 			dA=-(h22*g1 - h21 * g2) / det;
@@ -1656,14 +1658,14 @@ public class svm {
 				else
 					stepsize = stepsize / 2.0;
 			}
-			
+
 			if (stepsize < min_step)
 			{
 				svm.info("Line search fails in two-class probability estimates\n");
 				break;
 			}
 		}
-		
+
 		if (iter>=max_iter)
 			svm.info("Reaching maximal iterations in two-class probability estimates\n");
 		probAB[0]=A;probAB[1]=B;
@@ -1672,13 +1674,14 @@ public class svm {
 	private static double sigmoid_predict(double decision_value, double A, double B)
 	{
 		double fApB = decision_value*A+B;
+	// 1-p used later; avoid catastrophic cancellation
 		if (fApB >= 0)
 			return Math.exp(-fApB)/(1.0+Math.exp(-fApB));
 		else
 			return 1.0/(1+Math.exp(fApB)) ;
 	}
 
-	// Method 2 from the multiclass_prob paper by Wu, Lin, and Weng
+	// Method 2 from the multiclass_prob paper by Wu, Lin, and Weng to predict probabilities
 	private static void multiclass_probability(int k, double[][] r, double[] p)
 	{
 		int t,j;
@@ -1686,7 +1689,7 @@ public class svm {
 		double[][] Q=new double[k][k];
 		double[] Qp=new double[k];
 		double pQp, eps=0.005/k;
-	
+
 		for (t=0;t<k;t++)
 		{
 			p[t]=1.0/k;  // Valid if k = 1
@@ -1721,7 +1724,7 @@ public class svm {
 					max_error=error;
 			}
 			if (max_error<eps) break;
-		
+
 			for (t=0;t<k;t++)
 			{
 				double diff=(-Qp[t]+pQp)/Q[t][t];
@@ -1738,7 +1741,7 @@ public class svm {
 			svm.info("Exceeds max_iter in multiclass_prob\n");
 	}
 
-	// Cross-validation decision values for probability estimates
+	// Using cross-validation decision values to get parameters for SVC probability estimates
 	private static void svm_binary_svc_probability(svm_problem prob, svm_parameter param, double Cp, double Cn, double[] probAB)
 	{
 		int i;
@@ -1763,7 +1766,7 @@ public class svm {
 			subprob.l = prob.l-(end-begin);
 			subprob.x = new svm_node[subprob.l][];
 			subprob.y = new double[subprob.l];
-			
+
 			k=0;
 			for(j=0;j<begin;j++)
 			{
@@ -1783,7 +1786,7 @@ public class svm {
 					p_count++;
 				else
 					n_count++;
-			
+
 			if(p_count==0 && n_count==0)
 				for(j=begin;j<end;j++)
 					dec_values[perm[j]] = 0;
@@ -1813,13 +1816,82 @@ public class svm {
 					dec_values[perm[j]]=dec_value[0];
 					// ensure +1 -1 order; reason not using CV subroutine
 					dec_values[perm[j]] *= submodel.label[0];
-				}		
+				}
 			}
-		}		
+		}
 		sigmoid_train(prob.l,dec_values,prob.y,probAB);
 	}
 
-	// Return parameter of a Laplace distribution 
+	// Binning method from the oneclass_prob paper by Que and Lin to predict the probability as a normal instance (i.e., not an outlier)
+	private static double predict_one_class_probability(svm_model model, double dec_value)
+	{
+		double prob_estimate = 0.0;
+		int nr_marks = 10;
+
+		if(dec_value < model.prob_density_marks[0])
+			prob_estimate = 0.0001;
+		else if(dec_value > model.prob_density_marks[nr_marks-1])
+			prob_estimate = 0.999;
+		else
+		{
+			for(int i=0;i<nr_marks;i++)
+				if(dec_value < model.prob_density_marks[i])
+				{
+					prob_estimate = (double)i/nr_marks;
+					break;
+				}
+		}
+		return prob_estimate;
+	}
+
+	// Get parameters for one-class SVM probability estimates
+	private static int svm_one_class_probability(svm_problem prob, svm_model model, double[] prob_density_marks)
+	{
+		double[] dec_values = new double[prob.l];
+		double[] pred_results = new double[prob.l];
+		int ret = 0;
+		int nr_marks = 10;
+
+		for(int i=0;i<prob.l;i++)
+		{
+			double[] dec_value = new double[1];
+			pred_results[i] = svm_predict_values(model,prob.x[i],dec_value);
+			dec_values[i] = dec_value[0];
+		}
+		Arrays.sort(dec_values);
+
+		int neg_counter=0;
+		for(int i=0;i<prob.l;i++)
+			if(dec_values[i]>=0)
+			{
+				neg_counter = i;
+				break;
+			}
+
+		int pos_counter = prob.l-neg_counter;
+		if(neg_counter<nr_marks/2 || pos_counter<nr_marks/2)
+		{
+			System.err.print("WARNING: number of positive or negative decision values <"+nr_marks/2+"; too few to do a probability estimation.\n");
+			ret = -1;
+		}
+		else
+		{
+			// Binning by density
+			double[] tmp_marks = new double[nr_marks+1];
+			int mid = nr_marks/2;
+			for(int i=0;i<mid;i++)
+				tmp_marks[i] = dec_values[i*neg_counter/mid];
+			tmp_marks[mid] = 0;
+			for(int i=mid+1;i<nr_marks+1;i++)
+				tmp_marks[i] = dec_values[neg_counter-1+(i-mid)*pos_counter/mid];
+
+			for(int i=0;i<nr_marks;i++)
+				prob_density_marks[i] = (tmp_marks[i]+tmp_marks[i+1])/2;
+		}
+		return ret;
+	}
+
+	// Return parameter of a Laplace distribution
 	private static double svm_svr_probability(svm_problem prob, svm_parameter param)
 	{
 		int i;
@@ -1834,15 +1906,15 @@ public class svm {
 		{
 			ymv[i]=prob.y[i]-ymv[i];
 			mae += Math.abs(ymv[i]);
-		}		
+		}
 		mae /= prob.l;
 		double std=Math.sqrt(2*mae*mae);
 		int count=0;
 		mae=0;
 		for(i=0;i<prob.l;i++)
-			if (Math.abs(ymv[i]) > 5*std) 
+			if (Math.abs(ymv[i]) > 5*std)
 				count=count+1;
-			else 
+			else
 				mae+=Math.abs(ymv[i]);
 		mae /= (prob.l-count);
 		svm.info("Prob. model for test data: target value = predicted value + z,\nz: Laplace distribution e^(-|z|/sigma)/(2sigma),sigma="+mae+"\n");
@@ -1884,7 +1956,7 @@ public class svm {
 					label = new_data;
 					new_data = new int[max_nr_class];
 					System.arraycopy(count,0,new_data,0,count.length);
-					count = new_data;					
+					count = new_data;
 				}
 				label[nr_class] = this_label;
 				count[nr_class] = 1;
@@ -1893,8 +1965,8 @@ public class svm {
 		}
 
 		//
-		// Labels are ordered by their first occurrence in the training set. 
-		// However, for two-class sets with -1/+1 labels and -1 appears first, 
+		// Labels are ordered by their first occurrence in the training set.
+		// However, for two-class sets with -1/+1 labels and -1 appears first,
 		// we swap labels to ensure that internally the binary SVM has positive data corresponding to the +1 instances.
 		//
 		if (nr_class == 2 && label[0] == -1 && label[1] == +1)
@@ -1946,16 +2018,9 @@ public class svm {
 			model.label = null;
 			model.nSV = null;
 			model.probA = null; model.probB = null;
+			model.prob_density_marks = null;
 			model.sv_coef = new double[1][];
 
-			if(param.probability == 1 &&
-			   (param.svm_type == svm_parameter.EPSILON_SVR ||
-			    param.svm_type == svm_parameter.NU_SVR))
-			{
-				model.probA = new double[1];
-				model.probA[0] = svm_svr_probability(prob,param);
-			}
-
 			decision_function f = svm_train_one(prob,param,0,0);
 			model.rho = new double[1];
 			model.rho[0] = f.rho;
@@ -1977,6 +2042,26 @@ public class svm {
 					model.sv_indices[j] = i+1;
 					++j;
 				}
+
+			if(param.probability == 1 &&
+			   (param.svm_type == svm_parameter.EPSILON_SVR ||
+			    param.svm_type == svm_parameter.NU_SVR))
+			{
+				model.probA = new double[1];
+				model.probA[0] = svm_svr_probability(prob,param);
+			}
+			else if(param.probability == 1 && param.svm_type == svm_parameter.ONE_CLASS)
+			{
+				int nr_marks = 10;
+				double[] prob_density_marks = new double[nr_marks];
+
+				if(svm_one_class_probability(prob,model,prob_density_marks) == 0)
+				{
+					model.prob_density_marks = new double[nr_marks];
+					for(i=0;i<nr_marks;i++)
+						model.prob_density_marks[i] = prob_density_marks[i];
+				}
+			}
 		}
 		else
 		{
@@ -1985,19 +2070,19 @@ public class svm {
 			int[] tmp_nr_class = new int[1];
 			int[][] tmp_label = new int[1][];
 			int[][] tmp_start = new int[1][];
-			int[][] tmp_count = new int[1][];			
+			int[][] tmp_count = new int[1][];
 			int[] perm = new int[l];
 
 			// group training data of the same class
 			svm_group_classes(prob,tmp_nr_class,tmp_label,tmp_start,tmp_count,perm);
-			int nr_class = tmp_nr_class[0];			
+			int nr_class = tmp_nr_class[0];
 			int[] label = tmp_label[0];
 			int[] start = tmp_start[0];
 			int[] count = tmp_count[0];
- 			
-			if(nr_class == 1) 
+
+			if(nr_class == 1)
 				svm.info("WARNING: training data in only one class. See README for details.\n");
-			
+
 			svm_node[][] x = new svm_node[l][];
 			int i;
 			for(i=0;i<l;i++)
@@ -2101,8 +2186,9 @@ public class svm {
 				model.probA=null;
 				model.probB=null;
 			}
+			model.prob_density_marks = null;	// for one-class SVM probabilistic outputs only
 
-			int nnz = 0;
+			int total_sv = 0;
 			int[] nz_count = new int[nr_class];
 			model.nSV = new int[nr_class];
 			for(i=0;i<nr_class;i++)
@@ -2112,17 +2198,17 @@ public class svm {
 					if(nonzero[start[i]+j])
 					{
 						++nSV;
-						++nnz;
+						++total_sv;
 					}
 				model.nSV[i] = nSV;
 				nz_count[i] = nSV;
 			}
 
-			svm.info("Total nSV = "+nnz+"\n");
+			svm.info("Total nSV = "+total_sv+"\n");
 
-			model.l = nnz;
-			model.SV = new svm_node[nnz][];
-			model.sv_indices = new int[nnz];
+			model.l = total_sv;
+			model.SV = new svm_node[total_sv][];
+			model.sv_indices = new int[total_sv];
 			p = 0;
 			for(i=0;i<l;i++)
 				if(nonzero[i])
@@ -2138,7 +2224,7 @@ public class svm {
 
 			model.sv_coef = new double[nr_class-1][];
 			for(i=0;i<nr_class-1;i++)
-				model.sv_coef[i] = new double[nnz];
+				model.sv_coef[i] = new double[total_sv];
 
 			p = 0;
 			for(i=0;i<nr_class;i++)
@@ -2167,7 +2253,7 @@ public class svm {
 		}
 		return model;
 	}
-	
+
 	// Stratified cross validation
 	public static void svm_cross_validation(svm_problem prob, svm_parameter param, int nr_fold, double[] target)
 	{
@@ -2175,7 +2261,7 @@ public class svm {
 		int[] fold_start = new int[nr_fold+1];
 		int l = prob.l;
 		int[] perm = new int[l];
-		
+
 		// stratified cv may not give leave-one-out rate
 		// Each class to l folds -> some folds may have zero elements
 		if((param.svm_type == svm_parameter.C_SVC ||
@@ -2190,7 +2276,7 @@ public class svm {
 
 			int nr_class = tmp_nr_class[0];
 			int[] start = tmp_start[0];
-			int[] count = tmp_count[0];		
+			int[] count = tmp_count[0];
 
 			// random shuffle and then data grouped by fold using the array perm
 			int[] fold_count = new int[nr_fold];
@@ -2343,7 +2429,7 @@ public class svm {
 		{
 			int nr_class = model.nr_class;
 			int l = model.l;
-		
+
 			double[] kvalue = new double[l];
 			for(i=0;i<l;i++)
 				kvalue[i] = Kernel.k_function(x,model.SV[i],model.param);
@@ -2366,7 +2452,7 @@ public class svm {
 					int sj = start[j];
 					int ci = model.nSV[i];
 					int cj = model.nSV[j];
-				
+
 					int k;
 					double[] coef1 = model.sv_coef[j-1];
 					double[] coef2 = model.sv_coef[i];
@@ -2375,7 +2461,7 @@ public class svm {
 					for(k=0;k<cj;k++)
 						sum += coef2[sj+k] * kvalue[sj+k];
 					sum -= model.rho[p];
-					dec_values[p] = sum;					
+					dec_values[p] = sum;
 
 					if(dec_values[p] > 0)
 						++vote[i];
@@ -2419,7 +2505,7 @@ public class svm {
 
 			double min_prob=1e-7;
 			double[][] pairwise_prob=new double[nr_class][nr_class];
-			
+
 			int k=0;
 			for(i=0;i<nr_class;i++)
 				for(int j=i+1;j<nr_class;j++)
@@ -2442,7 +2528,15 @@ public class svm {
 					prob_max_idx = i;
 			return model.label[prob_max_idx];
 		}
-		else 
+		else if(model.param.svm_type == svm_parameter.ONE_CLASS && model.prob_density_marks!=null)
+		{
+			double[] dec_value = new double[1];
+			double pred_result = svm_predict_values(model,x,dec_value);
+			prob_estimates[0] = predict_one_class_probability(model,dec_value[0]);
+			prob_estimates[1] = 1-prob_estimates[0];
+			return pred_result;
+		}
+		else
 			return svm_predict(model, x);
 	}
 
@@ -2481,14 +2575,14 @@ public class svm {
 		int l = model.l;
 		fp.writeBytes("nr_class "+nr_class+"\n");
 		fp.writeBytes("total_sv "+l+"\n");
-	
+
 		{
 			fp.writeBytes("rho");
 			for(int i=0;i<nr_class*(nr_class-1)/2;i++)
 				fp.writeBytes(" "+model.rho[i]);
 			fp.writeBytes("\n");
 		}
-	
+
 		if(model.label != null)
 		{
 			fp.writeBytes("label");
@@ -2504,13 +2598,21 @@ public class svm {
 				fp.writeBytes(" "+model.probA[i]);
 			fp.writeBytes("\n");
 		}
-		if(model.probB != null) 
+		if(model.probB != null)
 		{
 			fp.writeBytes("probB");
 			for(int i=0;i<nr_class*(nr_class-1)/2;i++)
 				fp.writeBytes(" "+model.probB[i]);
 			fp.writeBytes("\n");
 		}
+		if(model.prob_density_marks != null)
+		{
+			fp.writeBytes("prob_density_marks");
+			int nr_marks = 10;
+			for(int i=0;i<nr_marks;i++)
+				fp.writeBytes(" "+model.prob_density_marks[i]);
+			fp.writeBytes("\n");
+		}
 
 		if(model.nSV != null)
 		{
@@ -2532,7 +2634,7 @@ public class svm {
 			svm_node[] p = SV[i];
 			if(param.kernel_type == svm_parameter.PRECOMPUTED)
 				fp.writeBytes("0:"+(int)(p[0].value));
-			else	
+			else
 				for(int j=0;j<p.length;j++)
 					fp.writeBytes(p[j].index+":"+p[j].value+" ");
 			fp.writeBytes("\n");
@@ -2555,7 +2657,7 @@ public class svm {
 	{
 		svm_parameter param = new svm_parameter();
 		model.param = param;
-		// parameters for training only won't be assigned, but arrays are assigned as NULL for safety
+		// parameters for training only won't be assigned, but arrays are assigned as null for safety
 		param.nr_weight = 0;
 		param.weight_label = null;
 		param.weight = null;
@@ -2625,7 +2727,7 @@ public class svm {
 					model.label = new int[n];
 					StringTokenizer st = new StringTokenizer(arg);
 					for(int i=0;i<n;i++)
-						model.label[i] = atoi(st.nextToken());					
+						model.label[i] = atoi(st.nextToken());
 				}
 				else if(cmd.startsWith("probA"))
 				{
@@ -2633,7 +2735,7 @@ public class svm {
 					model.probA = new double[n];
 					StringTokenizer st = new StringTokenizer(arg);
 					for(int i=0;i<n;i++)
-						model.probA[i] = atof(st.nextToken());					
+						model.probA[i] = atof(st.nextToken());
 				}
 				else if(cmd.startsWith("probB"))
 				{
@@ -2641,7 +2743,15 @@ public class svm {
 					model.probB = new double[n];
 					StringTokenizer st = new StringTokenizer(arg);
 					for(int i=0;i<n;i++)
-						model.probB[i] = atof(st.nextToken());					
+						model.probB[i] = atof(st.nextToken());
+				}
+				else if(cmd.startsWith("prob_density_marks"))
+				{
+					int n = 10;	// nr_marks
+					model.prob_density_marks = new double[n];
+					StringTokenizer st = new StringTokenizer(arg);
+					for(int i=0;i<n;i++)
+						model.prob_density_marks[i] = atof(st.nextToken());
 				}
 				else if(cmd.startsWith("nr_sv"))
 				{
@@ -2682,10 +2792,12 @@ public class svm {
 		model.rho = null;
 		model.probA = null;
 		model.probB = null;
+		model.prob_density_marks = null;
 		model.label = null;
 		model.nSV = null;
 
-		if (read_model_header(fp, model) == false)
+		// read header
+		if (!read_model_header(fp, model))
 		{
 			System.err.print("ERROR: failed to read model\n");
 			return null;
@@ -2732,7 +2844,7 @@ public class svm {
 		return "unknown svm type";
 
 		// kernel_type, degree
-	
+
 		int kernel_type = param.kernel_type;
 		if(kernel_type != svm_parameter.LINEAR &&
 		   kernel_type != svm_parameter.POLY &&
@@ -2741,10 +2853,13 @@ public class svm {
 		   kernel_type != svm_parameter.PRECOMPUTED)
 			return "unknown kernel type";
 
-		if(param.gamma < 0)
+		if((kernel_type == svm_parameter.POLY ||
+		   kernel_type == svm_parameter.RBF ||
+		   kernel_type == svm_parameter.SIGMOID) &&
+		   param.gamma < 0)
 			return "gamma < 0";
 
-		if(param.degree < 0)
+		if(kernel_type == svm_parameter.POLY && param.degree < 0)
 			return "degree of polynomial kernel < 0";
 
 		// cache_size,eps,C,nu,p,shrinking
@@ -2779,12 +2894,8 @@ public class svm {
 		   param.probability != 1)
 			return "probability != 0 and probability != 1";
 
-		if(param.probability == 1 &&
-		   svm_type == svm_parameter.ONE_CLASS)
-			return "one-class SVM probability output not supported yet";
-		
 		// check whether nu-svc is feasible
-	
+
 		if(svm_type == svm_parameter.NU_SVC)
 		{
 			int l = prob.l;
@@ -2813,7 +2924,7 @@ public class svm {
 						int[] new_data = new int[max_nr_class];
 						System.arraycopy(label,0,new_data,0,label.length);
 						label = new_data;
-						
+
 						new_data = new int[max_nr_class];
 						System.arraycopy(count,0,new_data,0,count.length);
 						count = new_data;
@@ -2843,6 +2954,7 @@ public class svm {
 	{
 		if (((model.param.svm_type == svm_parameter.C_SVC || model.param.svm_type == svm_parameter.NU_SVC) &&
 		model.probA!=null && model.probB!=null) ||
+		(model.param.svm_type == svm_parameter.ONE_CLASS && model.prob_density_marks!=null) ||
 		((model.param.svm_type == svm_parameter.EPSILON_SVR || model.param.svm_type == svm_parameter.NU_SVR) &&
 		 model.probA!=null))
 			return 1;
@@ -2854,7 +2966,7 @@ public class svm {
 	{
 		if (print_func == null)
 			svm_print_string = svm_print_stdout;
-		else 
+		else
 			svm_print_string = print_func;
 	}
 }
